<?php // Please do not edit this file. If you do, save as UTF-8 without BOM!

/* Copyright 2019-2022 Martin Janecke <mouse@prlbr.de>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

// declare (strict_types = 1); // only during development

namespace prlbr\mouse;

require_once (__DIR__ . '/word.php');


/* This class finds documents for a given search string by searching a database
 * build with the related feeder class.
 */
final class finder {

// This class recognizes two ways that a search term can match a word in a
// document. The first type is a full match. Consider the search term 'roo'.
// We have a full match, if a document contains the word 'roo'. The second
// type is a partial match, which occurs where a document contains a word
// (e.g. 'room') that begins with the complete search term or ends with the
// complete search term (e.g. 'kangaroo').
// full_weight defines the importance of a full match compared to a partial
// match. A value of 2 means that a full match has twice the weight of a partial
// match and leads to a higher ranking in the search results.
private const full_weight = 2;


/* Properties
 */

// SQLite3 object used to connect to the actual database
private object $db;


/* Methods
 */

/* The constructor connects to the SQLite database file located at $path. Using
 * journal_mode WAL improves concurrency, see https://www.sqlite.org/wal.html .
 */
public function __construct (
	string $path
)
{
	$this->db = new \SQLite3 ($path, SQLITE3_OPEN_READWRITE);
	$this->db->busyTimeout (10000); // 10 seconds
	$this->db->exec ('PRAGMA journal_mode=WAL;');
}


/* The destructor closes the connection to the database, so we do not do that
 * explicitely anywhere else.
 */
public function __destruct (
)
{
	$this->db->close ();
}


/* Expects a dot-separated $string (possibly with trailing dot) of occurrence
 * definitions. An occurrence definition consists of a document id, a colon and
 * a comma-separated sequence of relative word places. The entries in the column
 * 'docs' in the database tables 'title' and 'content' are such strings. Returns
 * an array whose keys are the document IDs and whose values are arrays of
 * corresponding absolute places. Example: the function returns
 * [1 => [60, 64, 125], 3 => [7, 38, 39]] for a given $string
 * "1:60,4,61.3:7,31,1.". Every document ID should only occur once.
 */
private static function docslist (
	string $string
):	array
{
	$docslist = [];
	foreach (explode ('.', rtrim ($string, '.')) as $definition):
		[$doc, $relative] = explode (':', $definition);
		$docslist[$doc] = self::partialsums ($relative);
	endforeach;
	return $docslist;
}


/* Works similar to the function docslist (see there), but returns an array
 * whose keys are the document IDs and whose values are just the number of
 * occurrences in the respective document instead of an array of the places of
 * occurrence. Mind that the same document IDs may occur multiple times here,
 * because $string can be a concatenation of results for different words.
 */
private static function docscount (
	string $string
):	array
{
	$docscount = [];
	foreach (explode ('.', rtrim ($string, '.')) as $definition):
		$doc = substr ($definition, 0, strpos ($definition, ':'));
		$count = substr_count ($definition, ',') + 1;
		if (isset ($docscount[$doc])):
			$docscount[$doc] += $count;
		else:
			$docscount[$doc] = $count;
		endif;
	endforeach;
	return $docscount;
}


/* Searches for a given $searchstring which can contain multiple terms separated
 * by space. Terms and phrases can be enclosed in quotes to only look for exact
 * matches; e.g. "roo" finds only the word roo, but not the word room.
 * Returns an array of hits whose keys are the URLs of found documents and the
 * values are arrays of the form [document title, score]. A higher score, which
 * is calculated by this function, implies a better match.
 */
public function find (
	string $searchstring,
	string $charset = 'UTF-8'
):	array
{
	// parse the search string into an array of search terms and find matching
	// documents; the keys of the $found array are document IDs, the
	// corresponding values contain information to be used for ranking
	$found = $this->find_terms (word::parse ($searchstring, $charset));
	if (empty ($found)):
		return [];
	endif;

	// compute the average number of words in a document from all documents in
	// the database; this number is used in the ranking algorithm
	if ($row = $this->db->querySingle ("SELECT SUM(size) AS s, COUNT(*) AS c FROM document", true)):
		$avrg = $row['c']? $row['s'] / $row['c']: 0;
	else:
		trigger_error (__METHOD__ . " could not get total/count");
		$avrg = 0;
	endif;

	// use the IDs of the found documents to retrieve meta data about the found
	// documents from the database, i.e. url, title, number of words
	$docs = implode (', ', array_keys ($found)); // document IDs
	$rows = [];
	$size = []; // number of words in each document
	$result = $this->db->query ("SELECT * FROM document WHERE id IN ($docs)");
	while ($row = $result->fetchArray (SQLITE3_ASSOC)):
		$rows[$row['id']] = $row;
	endwhile;
	unset ($result);

	// compute the score of each document found for the given search
	$score = [];
	foreach ($found as $doc => $words):
		$docscore = 0;
		foreach ($words as $stats):
			if ($t = $stats['title']):
				$docscore += (self::full_weight * $t['full'] + $t['part'])/(self::full_weight * $t['fsum'] + $t['psum']);
			endif;
			if ($c = $stats['content']):
				$docscore += (self::full_weight * $c['full'] + $c['part'])/(self::full_weight * $c['fsum'] + $c['psum']) * sqrt ($avrg / $rows[$doc]['size']);
			endif;
		endforeach;
		$rows[$doc]['score'] = $score[$doc] = $docscore;
	endforeach;
	unset ($found);

	// sort the array of document IDs and their corresponding scores by the
	// scores, descending
	arsort ($score, SORT_NUMERIC);

	// build the array of ranked results using the sorted document IDs
	$ranked = [];
	foreach ($score as $doc => $docscore):
		$ranked[] = $rows[$doc];
	endforeach;

	return $ranked;
}


/* Searches for a given exact $phrase in the database $table 'title' or
 * 'content'. The $phrase must be an array of words; an array with a single word
 * is fine. Returns an array whose keys are IDs of the documents which contain
 * the phrase and the values are arrays with statistical information.
 */
private function find_phrase_in_table (
	string $table, // must be 'title' or 'content'
	array $phrase  // an array of words
):	array
{
	// store all occurrences of each word from $phrase in the array $words
	$words = [];
	$select = $this->db->prepare ("SELECT docs FROM $table WHERE word = ?");
	$select->bindParam (1, $word, SQLITE3_TEXT);
	foreach ($phrase as $word):
		$result = $select->execute ();
		$row = $result->fetchArray (SQLITE3_NUM);
		$words[] = empty ($row)? []: self::docslist ($row[0]);
		$select->reset ();
	endforeach;

	// return empty result, if none of the words in the phrase was found or the
	// first word of the phrase was not found anywhere in the database
	if (empty ($words) or empty ($words[0])):
		return [];
	endif;

	// only retain those occurrences of the first word from the phrase, which
	// are in documents that also contain all other words from the phrase
	$length = count ($phrase);
	if ($length > 1):
		$words[0] = array_intersect_key ($words[0], ...$words);
	endif;

	$total = 0; // for total number of occurrences of the phrase in documents
	$hits = []; // for information about documents that contain the phrase

	// check for every document that contains the first word from the phrase and
	// for every occurrence of the first word in it whether the words that
	// follow in the phrase also follow in the document; it's a hit when they do
	foreach ($words[0] as $doc => $places):
		foreach ($places as $start):
			for ($i = 1; $i < $length; $i++):
				if (!in_array ($start + $i, $words[$i][$doc])):
					continue 2;
				endif;
			endfor;
			// it's a hit, iff we get here despite the 'continue 2' above
			$total++;
			if (isset ($hits[$doc])):
				$hits[$doc]['full']++;
			else:
				$hits[$doc] = [
					'full' => 1,
					'part' => 0,
					'fsum' => &$total, // by reference to update automatically
					'psum' => 0
				];
			endif;
		endforeach;
	endforeach;
	
	return $hits;
}


/* Searches for an array of $terms in the title and contents. Each element of
 * $terms is either of type string and a word or of type array and a phrase,
 * which can contain multiple words.
 * Returns an array whose keys are IDs of the documents which contain all search
 * terms and whose values are arrays with statistical information about each
 * term, how often it occurs where in the document and how often in the whole
 * database.
 */
private function find_terms (
	array $terms
):	array
{
	// re-index the array from 0
	$terms = array_values ($terms);

	// find titles and contents that contain at least one of the search terms
	$t_hits = $this->find_terms_in_table ('title', $terms);
	$c_hits = $this->find_terms_in_table ('content', $terms);
		
	// get the IDs of documents that contain at least one search term
	$t_docs = array_keys ($t_hits);
	$c_docs = array_keys ($c_hits);

	// make $i_docs contain those docs that contain search terms in both the
	// title and content, $t_docs those with matches only in the title and
	// $c_docs those with matches only in the content
	$i_docs = array_intersect ($t_docs, $c_docs);
	$t_docs = array_diff ($t_docs, $i_docs);
	$c_docs = array_diff ($c_docs, $i_docs);

	// remove partial hits, i.e. those docs where not all terms were found
	$termcount = count ($terms);
	foreach ($i_docs as $n => $doc):
		if (!empty (array_diff_key ($terms, $t_hits[$doc], $c_hits[$doc]))):
			unset ($t_hits[$doc], $c_hits[$doc], $i_docs[$n]);
		endif;
	endforeach;
	foreach ($t_docs as $n => $doc):
		if (count ($t_hits[$doc]) < $termcount):
			unset ($t_hits[$doc], $t_docs[$n]);
		endif;
	endforeach;
	foreach ($c_docs as $n => $doc):
		if (count ($c_hits[$doc]) < $termcount):
			unset ($c_hits[$doc], $c_docs[$n]);
		endif;
	endforeach;

	// prepare an array that stores all the hits
	$found = [];
	foreach (array_merge ($i_docs, $t_docs, $c_docs) as $doc):
		$found[$doc] = array_fill (0, $termcount, ['title' => false, 'content' => false]);
	endforeach;

	// fill the array with detailed information about all hits
	foreach ($i_docs as $doc):
		foreach ($t_hits[$doc] as $key => $value):
			$found[$doc][$key]['title'] = $value;
		endforeach;
		foreach ($c_hits[$doc] as $key => $value):
			$found[$doc][$key]['content'] = $value;
		endforeach;
	endforeach;
	foreach ($t_docs as $doc):
		foreach ($t_hits[$doc] as $key => $value):
			$found[$doc][$key]['title'] = $value;
		endforeach;
	endforeach;
	foreach ($c_docs as $doc):
		foreach ($c_hits[$doc] as $key => $value):
			$found[$doc][$key]['content'] = $value;
		endforeach;
	endforeach;

	return $found;
}


/* Searches for an array of $terms in the database $table 'title' or 'content'.
 * Each element of $terms is either of type string and a word or of type array
 * and a phrase, which can contain multiple words.
 * Returns an array whose keys are IDs of the documents which contain at least
 * one search term and whose values are arrays with statistical information
 * about each term that is found in the document.
 */
private function find_terms_in_table (
	string $table, // must be 'title' or 'content'
	array $terms
):	array
{
	// we'll search for full matches and partial matches separately, because
	// these are weighted differently; note that there can be only one result
	// row per request for a full match, but multiple partial matches
	$full = $this->db->prepare ("SELECT docs FROM $table WHERE word = ?");
	$part = $this->db->prepare ("SELECT docs FROM $table WHERE word LIKE ? OR word LIKE ?");
	$full->bindParam (1, $word, SQLITE3_TEXT);	
	$part->bindParam (1, $word_, SQLITE3_TEXT);	
	$part->bindParam (2, $_word, SQLITE3_TEXT);	

	$hits = []; // for information about documents that contain the phrase

	foreach ($terms as $key => $term):
		if (is_array ($term)): // phrases are handled by a separate function
			foreach ($this->find_phrase_in_table ($table, $term) as $doc => $hit):
				$hits[$doc][$key] = $hit;
			endforeach;
		else: // word
			$word = $term;
			$word_ = $term . '_%';
			$_word = '%_' . $term;

			$result = $full->execute ();
			$row = $result->fetchArray (SQLITE3_NUM);
			$full->reset ();
			$fulldocs = empty ($row)? []: self::docscount ($row[0]);
			$fullsum = array_sum ($fulldocs);

			$result = $part->execute ();			
			$rows = [];	while ($row = $result->fetchArray (SQLITE3_NUM)) $rows[] = $row[0];
			$part->reset ();
			$partdocs = empty ($rows)? []: self::docscount (implode ($rows));
			$partsum = array_sum ($partdocs);

			foreach ($fulldocs as $doc => $count):
				$hits[$doc][$key] = [
					'full' => $count,
					'part' => 0,
					'fsum' => $fullsum,
					'psum' => $partsum
				];
			endforeach;

			foreach ($partdocs as $doc => $count):
				if (isset ($hits[$doc][$key])): // may be set in foreach ($fulldocs)
					$hits[$doc][$key]['part'] = $count;
				else:
					$hits[$doc][$key] = [
						'full' => 0,
						'part' => $count,
						'fsum' => $fullsum,
						'psum' => $partsum
					];
				endif;
			endforeach;
		endif;
	endforeach;

	return $hits;
}


/* Returns information about the documents identified by the $requested URLs.
 * The numeric keys of the $requested array are expected to be sorted ascending.
 */
public function get_by_urls (
	array $requested
):	array
{
	if (empty ($requested)):
		return [];
	endif;

	// Fetch the rows for the requested URLs from the table 'document':
	$urls = implode ("','", array_map ('\SQLite3::escapeString', $requested));
	$result = $this->db->query ("SELECT * FROM document WHERE url IN ('$urls')");
	if ($result === false):
		return [];
	endif;

	// Create an array where keys and URLs are flipped so that we can then
	// ascribe the original keys based on the the URLs in the database response:
	$keys = array_flip ($requested);

	// Order the rows returned from the database like the requested URLs were
	// ordered. Mind that there is not necessarily a response for every
	// requested URL, since the respective document may not be in the database.
	$rows = [];
	while ($row = $result->fetchArray (SQLITE3_ASSOC)):
		$rows[$keys[$row['url']]] = $row;
	endwhile;
	ksort ($rows, SORT_NUMERIC);

	return $rows;
}


/* Returns information about $count random documents.
 */
public function get_random (
	int $count = 1
):	array
{
	if ($count < 1):
		return [];
	endif;

	// Fetch random rows from the table 'document':
	$result = $this->db->query ("SELECT * FROM document WHERE id IN (SELECT id FROM document ORDER BY RANDOM() LIMIT $count)");
	if ($result === false):
		return [];
	endif;

	$rows = [];
	while ($row = $result->fetchArray (SQLITE3_ASSOC)):
		$rows[] = $row;
	endwhile;
	return $rows;
}


/* Expects a $sequence of comma-separated integers and returns the partial sums
 * of the series as an array of ints. This is the inverse of feeder::distances.
 * See there for an example.
 */
private static function partialsums (
	string $sequence
):	array
{
	$integers = [];
	$previous = 0;
	foreach (explode (',', $sequence) as $distance):
		$integers[] = $previous += (int) $distance;
	endforeach;
	return $integers;
}

} // class finder

// The End