<?php // Please do not edit this file. If you do, save as UTF-8 without BOM!

/* Copyright 2021, 2022 Martin Janecke <mouse@prlbr.de>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

// declare (strict_types = 1); // only during development

namespace prlbr\mouse;

require_once __DIR__ . '/finder.php';


/* This class generates a news feed for readers to subscribe to. It currently
 * offers the Atom format but could also generate a JSON feed or RSS in the
 * future.
 */
class newsfeed {

// List of supported formats, used as file extension and must have an eponymous
// method that generates the output in the format:
private const formats = [
	'atom'
	// Note: consider supporting JSON feed (https://www.jsonfeed.org/) and/or
	// JF2 feed (https://www.w3.org/TR/jf2/). The feed could then also be used
	// by [x-mouse/news] itself insted of the current function json_export.
];

// TODO: change this when the version changes.
private const generator = 'Mouse/2';

// Number of recent webpages to be included in a feed. 15 is just a number that
// is neither huge nor too small. The idea is to have at least as many URLs as
// could get published during periodic checks by a client.
public const quantity = 15;


/* Properties
 */

// If this is not the empty string, an error described in the string occurred.
public string $error = '';

// Data about the homepage that serves pars pro toto as info for the whole feed,
// e.g. the homepage title and author are also taken as feed title and author.
private array $homepage = [];

// Information about the recent webpages found:
private array $entries = [];


/* Methods
 */

/* The constructor gathers data used for the news feed from a $textfile, which
 * contains a list of discovered webpages in the order of discovery and
 * timestamps of their discovery, and from an array of $databases which contain
 * up-to-date information about these webpages like title, author, description.
 * The primarily used database is the last one in $databases, earlier ones are
 * only used, if information could not be found in later ones.
 */
public function __construct (
	string $textfile,
	array $databases
)
{
	// Load the text file. It should be readable normally, so we do not check
	// that in advance, only in case reading does not give results.
	$file = @file ($textfile, FILE_IGNORE_NEW_LINES | FILE_SKIP_EMPTY_LINES);
	if (empty ($file)):
		$this->error = is_readable ($textfile)? 'empty news file': 'news file not readable';
		return;
	endif;

	// Extract records from the file to inquire about in the database: the first
	// item which is supposedly the home page that is used pars pro toto for the
	// general feed data plus a quantity of most recent found URLs. Each record
	// is an array where the first item is the URL, the second item is a
	// timestamp of its first dicovery and subsequent items are legacy debris.
	if (count ($file) > self::quantity):
		$records = [explode ("\t", $file[0])];
	else:
		$records = [];
	endif;
	foreach (array_slice ($file, -self::quantity) as $record):
		$records[] = explode ("\t", $record);
	endforeach;
	unset ($file);

	// Get data from the $databases. Usually there are two of them: a big but
	// older database and a newer, but not yet complete database.
	$urls = array_column ($records, 0);
	$news = [];
	while ($urls and $database = array_pop ($databases)):

		// Get fresh data about the URLs from the database to see whether the
		// pages still exist and to have up-to-date information about them:
		$db = new finder ($database);
		$hits = $db->get_by_urls ($urls);

		// Add the timestamp of the first discovery.
		// Make sure that all other properties needed by this class are set,
		// which is not always the case e.g. with Mouse 1.x legacy databases.
		foreach ($hits as $n => $hit):
			$news[$n]['url'] = $hit['url'] ?? '';
			$news[$n]['name'] = $hit['name'] ?? '';
			$news[$n]['desc'] = $hit['desc'] ?? '';
			$news[$n]['author'] = $hit['author'] ?? '';
			$news[$n]['timestamp'] = intval ($records[$n][1] ?? 0);
		endforeach;

		$urls = array_diff_key ($urls, $hits);
	endwhile;

	if (empty ($news)):
		$this->error = 'no hits in databases';
		return;
	endif;
	ksort ($news);

	// Separate the homepage data from the most recent pages. If there is just
	// the homepage, it is also kept as entry.
	$this->homepage = count ($news) > 1? array_shift ($news): reset ($news);
	$this->entries = array_values ($news);
}


/* Return the newsfeed as Atom feed. $hash is an ID to identify this feed, $self
 * is the URL of the Atom feed file. See:
 * http://www.atomenabled.org/developers/syndication/
 * https://datatracker.ietf.org/doc/html/rfc4287
 */
public function atom (
	string $hash,
	string $self = ''
):	string
{
	// We us the author of the homepage author, if not empty, as feed author.
	// A feed author is not required, but serves as default author for entries,
	// where an author is otherwise required.
	$feedauthor = self::escxml ($this->homepage['author']);

	// Generate the <entry> elements of the Atom feed:
	$entries = [];
	foreach ($this->entries as $entry):

		// Prepare data from the database for use in XML:
		$url = self::escxml ($entry['url']);
		$title = self::escxml ($entry['name']);
		$desc = self::escxml ($entry['desc']);
		$writer = self::escxml ($entry['author']);

		// Prepare the optional <summary>:
		$summary = $desc !== ''? "<summary>$desc</summary>": '';

		// <author> is only optional in <entry>, if there is an author for the
		// <feed>. We use U+1F42D MOUSE FACE if neither is available.
		if ($writer !== ''):
			$author = "<author><name>$writer</name></author>";
		elseif ($feedauthor !== ''):
			$author = '';
		else:
			$author = "<author><name>\xF0\x9F\x90\xAD</name></author>";
		endif;

		// Use the date of the first ever discovery for the required <updated>
		// element, since Mouse does not track actual updates.
		$updated = gmdate (DATE_ATOM, $entry['timestamp']);
		
		// Compile the entry:
		$entries[] = <<<MUS
			<entry>
				<id>$url</id>
				<title>$title</title>
				$author
				$summary
				<link href='$url' rel='alternate' />
				<updated>$updated</updated>
			</entry>
		MUS;
	endforeach;
	$entries = implode ("\n", $entries);

	// Generate the information about the whole feed now ...
	// It is taken pars pro toto from the home page.
	$home = self::escxml ($this->homepage['url']);
	$title = self::escxml ($this->homepage['name']);

	// The optional link to the feed itself helps people/software to subscribe
	// who just got a copy of the feed file, not its location.
	$self = $self === ''? '': "<link href='" . self::escxml ($self) . "' rel='self' type='application/atom+xml' />";

	// Prepare the optional feed <author> element:
	$author = $feedauthor === ''? '': "<author><name>$feedauthor</name></author>";

	// Use the current time when the feed is (re-)generated as update time for
	// the feed, since Mouse does not track actual updates to pages.
	$updated = gmdate (DATE_ATOM, time ());
	$generator = self::generator;

	// Compile and return the feed:
	$id = self::escxml ($hash);
	return <<<MUS
		<?xml version='1.0' encoding='UTF-8' ?\x3E
		<feed xmlns='http://www.w3.org/2005/Atom'>
			<id>$home#$id</id>
			<title>$title</title>
			$author
			<link href='$home' rel='alternate' type='text/html' />
			$self
			<generator>$generator</generator>
			<updated>$updated</updated>
		$entries
		</feed>
		MUS;
}


/* Escapes characters with a special meaning in XML so that a given $string can
 * be printed in an XML attribute.
 */
private static function escxml (
	string $string
):	string
{
	return htmlspecialchars ($string, ENT_QUOTES | ENT_SUBSTITUTE | ENT_XML1, 'UTF-8');
}


/* Exports entries to a JSON file.
 */
public function json_export (
	string $filename
):	void
{
	file_put_contents ($filename, json_encode ($this->entries, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT));
}


/* Saves the feed in all supported formats in a given $directory. The files'
 * basenames are compiled from $hash and an extension matching the format.
 * Returns true on success, false on failure.
 */
public function save (
	string $directory,
	string $hash // A-Z, a-z, 0-9
):	bool
{
	if ($this->error):
		return false;
	endif;

	// Try to create the directory, if it does not exist yet:
	if (!is_dir ($directory) and !@mkdir ($directory, 0777, true)):
		$this->error = "could not create feed directory '$directory'";
		return false;
	endif;

	// Save the feed in all supported formats:
	foreach (self::formats as $format):
		if (!@file_put_contents ("$directory/$hash.$format", $this->$format ($hash))):
			$this->error ("could not save '$hash.$format' feed in '$directory'");
			return false;
		endif;
	endforeach;

	// Everything worked fine, if we got here.
	return true;
}

} // class newsfeed

// The End