<?php // Please do not edit this file. If you do, save as UTF-8 without BOM!

/* Copyright 2019-2022 Martin Janecke <mouse@prlbr.de>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

// declare (strict_types = 1); // only during development

namespace prlbr\mouse;

require_once __DIR__ . '/word.php';


/* This class fetches text documents over the internet and prepares them for
 * inclusion in a search database. It also finds links to new documents.
 * HTML meta robots tags and HTTP X-Robots headers as well as further signals to
 * bots are respected. robots.txt is not taken into account by this class
 * internally, but it offers a method to fetch robots.txt so that it can be
 * respected outside this class.
 */
class retriever {

// We send an Accept header with these types hoping that servers will respond
// with status 406 if the resource does not match the type, but we can not rely
// on this. So we will also check the type of the response ourselves and discard
// unacceptable responses.
private const accept_types = [
	'text/html', 'text/plain', 'application/xhtml+xml',
	// Note: SVGs, PDFs and VTTs could be indexed in the future.
	// Note: RSS, Atom feeds, XML sitemaps could be parsed for link discovery.
];

// These common file extensions are know to be used for content that is not
// indexable by this search engine, so links that lead to paths ending in these
// are not reported.
private const avoid_extensions = [
	'7z', 'avif', 'bmp', 'css', 'doc', 'docx', 'exe', 'flv', 'gif', 'gz',
	'heic', 'heif', 'ico', 'jpeg', 'jpg', 'jxl', 'js', 'json', 'mid', 'midi',
	'mov', 'mp3', 'mp4', 'mpg', 'ogg', 'ogv', 'pdf', 'png', 'ppt', 'pptx',
	'rar', 'svg', 'svgz', 'swf', 'tif', 'tiff', 'wav', 'webm', 'webp', 'xls',
	'xlsx', 'xml', 'zip',
];

// These link relations shall not be followed, for example because they signal
// non-indexable content like 'stylesheet', a redirect to the current document
// like 'shortlink' or explicitly say so like 'nofollow' (although search
// engines like Google interpret this relation a little different).
private const avoid_link_relations = [
	'icon', 'nofollow', 'shortlink', 'stylesheet',
];

// These non-self-closing elements in HTML will be padded with space characters
// to prevent them from collapsing two words to a single one when the tags are
// removed, so that for example '<h1>Sky</h1>The …' does not become 'SkyThe …'
// but ' Sky The …'.
private const elements_to_pad = [
	'address', 'article', 'aside', 'blockquote', 'body', 'caption', 'center',
	'cite', 'details', 'dialog', 'div', 'dd', 'dt', 'fieldset', 'figcaption',
	'figure', 'footer', 'form', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'header',
	'legend', 'li', 'main', 'menu', 'menuitem', 'nav', 'noscript', 'option',
	'p', 'pre', 'section', 'sub', 'sup', 'summary', 'rb', 'rt', 'td', 'th'
];

// Strip <style> and <script> and <template> because they are not text content.
// Remove <nav> because it is usually not relevant to this document but OTHER
// documents or simply repetitive. Remove <s> because it represents content that
// is no longer accurate or relevant. Remove <aside> because it is not the
// actual topic of the page, but advertisement, or repetitive.
private const elements_to_remove = [
	'aside', 'nav', 's', 'script', 'style', 'template'
];

// Strip <header> and <footer> elements from documents that were generated from
// Aneamal since they contain repetitive information to be found on all or many
// pages. We cannot remove them for non-Aneamal documents because the elements
// are ambiguous in HTML and can be relevant.
private const elements_to_remove_from_nml = [
	'footer', 'header'
];

// Remove elements with role=navigation and role=search because like <nav> it is
// usually not relevant to this document but OTHER documents or just repetitive.
private const elements_with_roles_to_remove = [
	'navigation', 'search'
];

// The self-closing elements <br> and <hr> are replaced with a space, so that
// words separated by them do not get merged when all tags are stripped.
private const elements_to_replace_with_space = [
	'br', 'hr'
];

// The following directives for robots meta tags and X-Robots-Tag headers take
// a value, from which they are separated by a colon. This syntax is ambiguous
// at the beginning of a X-Robots-Tag header: it can be confused with a user
// agent, to which subsequent directives are limited. We list known directives
// that could be confused in order not to confuse them for a user agent. Mind
// that Mouse may not interpet and adhere to these directives though! See:
// https://developers.google.com/search/docs/advanced/robots/robots_meta_tag#directives
private const parametric_directives = [
	'max-snippet', 'max-image-preview', 'max-video-preview', 'unavailable_after'
];

// Documents are capped at eight MiB. Larger files are unlikely to be
// indexable text anyway, but video files or executables instead.
private const maxsize = 8388608;

// The HTML <meta> author combined are capped at this many Unicode characters.
private const maxsize_author = 255;

// The HTML <meta> keywords combined are capped at this many Unicode characters.
private const maxsize_keywords = 510;

// robots.txt are capped at 500 KiB, an unrealistic big size.
private const maxsize_robots = 512000;

// HTML <meta> descriptions are capped at this many Unicode characters.
private const maxsize_snippet = 510;

// HTML <title>s are capped at this many Unicode characters.
private const maxsize_title = 255;

// URLs that have more bytes than this after stripping the fragment will not be
// processed. People should not use longer navigable URLs for practical reasons,
// even if it is technically possible. The restriction prevents following some
// erroneous redirects until technical limitations would be reached.
private const maxsize_url = 255;

// How long we are willing to invest in fetching a resource; 16 seconds means
// that we would be able to obtain maxsize bytes with 512 kiB/s.
private const timeout = 16;

// The search engine identifies as this when fetching documents, just like web
// browsers usually send data about their engine and the operating system.
// TODO: change this when the version changes.
private const user_agent = 'Mouse/2 (https://prlbr.de/mouse)';

// The search engine "feels" addressed when encountering a User-Agent field
// in robots.txt that starts with this (case-insenstive match, lowercase here).
private const user_agent_simplified = 'mouse';


/* Public properties
 */

// creator of the document or its main content respectively
public string $author = '';
// document description appropriate for use in a directory of pages
public string $description = '';
// keywords relevant to the page, list separated by ", "
public string $keywords = '';
// links from the document to other files that are in scope of the search engine
public array $links = [];
// document description (see above) or introductory paragraphs
public string $snippet = '';
// text content of the document that shall be indexed
public string $text = '';
// title from HTML <title> element, if available
public string $title = '';
// assume that the file is indexable and set to false on first counter argument
public bool $indexable = true;


/* Private properties
 */

// URL of the document that is fetched and analyzed
private string $url = '';
// base URL for links with relative URLs in the document
private string $base = '';
// array of strings with which URLs may start to be in the scope of this engine
private array $prefixes = [];
// array of response headers after fetching the document
private array $headers = [];
// array whose keys are robots directives from meta tags or x-robots-tag headers
private array $robots = [];


/* Methods
 */

/* The constructor triggers all the work on a document located at $url, i.e.
 * fetching and parsing it, extracting links and text contents. After the
 * constructor has done its work, the results can be obtained from public
 * properties of this class. $prefixes is an array of strings that define the
 * scope of this search engine: only links in the document with URLs that match
 * any of the given prefixes are reported in $this->links for further indexing.
 */
public function __construct (
	string $url,
	array $prefixes = []
)
{
	$this->url = word::strip_fragment ($url);
	$this->base = $this->url;

	foreach ($prefixes as $prefix):
		$prefix = word::normalize_url ($prefix);
		$this->prefixes[] = [$prefix, strlen ($prefix)];
	endforeach;

	// retrieve document
	list ($content, $headers) = self::fetch ($this->url);

	// Our job is finished here, if we do not have headers.
	if (empty ($headers)):
		$this->indexable = false;
		return;
	endif;

	// Empty content renders a document not indexable, but there may still be a
	// location header worth analyzing, so do not stop yet.
	if ($content === ''):
		$this->indexable = false;
	endif;

	// Get HTTP status code. Only the status code 200 marks an indexable
	// document.
	if ($space = strpos ($statusline = trim ($headers[0]), "\x20")):
		$status = (int) substr ($statusline, $space + 1);
	endif;
	if (!isset ($status) or $status !== 200):
		$this->indexable = false;
	endif;

	// turn header lines into key => value pairs
	foreach ($headers as $field):
		if (strpos ($field, ':')):
			list ($key, $value) = explode (':', $field, 2);
			$key = trim (strtolower ($key));
			$val = trim ($value);
			$this->headers[$key] = isset ($this->headers[$key])? $this->headers[$key] . "\n$val": $val;
		endif;
	endforeach;

	// Try to parse the X-Robots-Tag HTTP Header. "Try" because its syntax is
	// ambiguous. Take for example two complete headers:
	//   x-robots-tag: example-bot: none
    //   x-robots-tag: max-image-preview: none
    // In the first line "example-bot" is a user agent and "none" is a
    // directive without value whereas the second line is without user agent,
    // "max-image-preview" is a directive and "none" is its value.
	if (isset ($this->headers['x-robots-tag'])):
		foreach (explode ("\n", strtolower ($this->headers['x-robots-tag'])) as $line):
			// If a colon ":" comes before a comma "," and what is before the
			// colon is not a known directive that takes a value as parameter,
			// then we interpret everything before the colon as user agent. We
			// ignore lines with a user agent that does not match "mouse".
			if (
				$colon = strpos ($line, ':')
				and $colon < strpos ($line, ',')
				and !in_array (rtrim (substr ($line, 0, $colon)), self::parametric_directives, true)
			):
				if (substr ($line, 0, 5) === 'mouse'):
					$line = ltrim (substr ($line, $colon + 1));
				else:
					continue;
				endif;
			endif;
			foreach (explode (',', $line) as $value):
				$v = explode (':', $value, 2);
				$this->robots[trim ($v[0])] = trim ($v[1] ?? $v[0]);
			endforeach;
		endforeach;
	endif;
	if (isset ($this->robots['noindex']) or isset ($this->robots['none'])):
		$this->indexable = false;
	endif;

	// Note: consider interpreting a Link header with rel=canonical that differs
	// from the document's URL as sign not to index the document, similar to
	// <link rel='canonical'> in HTML

	// get the MIME type
	if (isset ($this->headers['content-type'])):
		list ($type) = explode (';', $this->headers['content-type']);
		$type = strtolower (rtrim ($type));
	else:
		$type = NULL;
	endif;

	// parse content of various formats
	if ($type === 'text/plain'):
		$this->parse_plain_text ($content);
	elseif ($type === 'text/html' or $type === 'application/xhtml+xml'):
		$this->parse_html ($content);
	else:
		$this->indexable = false;
	endif;
}


/* Returns the encoding of the document reported in the HTTP header. An empty
 * string is returned if no encoding has been found.
 */
private function get_charset_from_header (
):	string
{
	if ($start = stripos ($this->headers['content-type'], 'charset=')):
		return substr ($this->headers['content-type'], $start + 8, strcspn ($this->headers['content-type'], "()<>@,;:\\\"/[]?.=\t\r\n\x20", $start + 8));
	else:
		return '';
	endif;
}


/* Fetches a document over the internet and returns an array
 * [content, http_headers] where http_headers itself is an array. Redirects are
 * not followed, but the content of redirect pages as well as error pages such
 * as 404 Not Found is returned, because it may still contain useful information
 * like links that lead to relevant documents.
 */
private static function fetch (
	string $url
):	array
{
	if (in_array (strtolower ((string) ini_get ('allow_url_fopen')), ['1', 'on', 'yes', 'true'], true)):

		$content = trim ((string) file_get_contents ($url, false, stream_context_create ([
			'http' => [
				'header'		=> ['Accept: ' . implode (', ', self::accept_types)],
				'ignore_errors' => true,
				'max_redirects' => 0,
				'timeout'		=> self::timeout,
				'user_agent'	=> self::user_agent,
			]
		]), 0, self::maxsize));

		$headers = $http_response_header ?? [];

	else: // use cURL as fallback

		// Should we use CURLOPT_WRITEFUNCTION to get the first self::maxsize
		// bytes of documents that are too big, whereas CURLOPT_PROGRESSFUNCTION
		// leads to the whole document being discarded? It does not seem to be
		// important since it is only an edge case of a fallback solution.
		$progressfunction = function ($handle, $x, $downloaded, $y, $z) {
			return (int) $downloaded > self::maxsize;
		};

		$handle = curl_init ($url);
		curl_setopt ($handle, CURLOPT_FAILONERROR, false);
		curl_setopt ($handle, CURLOPT_FOLLOWLOCATION, false);
		curl_setopt ($handle, CURLOPT_HEADER, true);
		curl_setopt ($handle, CURLOPT_HTTPHEADER, ['Accept: ' . implode (', ', self::accept_types)]);
		curl_setopt ($handle, CURLOPT_NOBODY, false);
		curl_setopt ($handle, CURLOPT_NOPROGRESS, false);
		curl_setopt ($handle, CURLOPT_PROGRESSFUNCTION, $progressfunction);
		curl_setopt ($handle, CURLOPT_RETURNTRANSFER, true);
		curl_setopt ($handle, CURLOPT_TIMEOUT, self::timeout);
		curl_setopt ($handle, CURLOPT_USERAGENT, self::user_agent);
		$response = curl_exec ($handle);
		$errno = curl_errno ($handle) and $error = curl_error ($handle);
		curl_close ($handle);

		if ($errno):
			if ($errno === 42): // document too big
				return ['', []];
			else:
				trigger_error (__METHOD__ . " encountered error $errno: $error");
			endif;
		endif;

		// normalize line endings and split content into headers and body
		$response = str_replace (array ("\r\n", "\r"), "\n", $response);
		if (strpos ($response, "\n\n")):
			list ($headers, $content) = explode ("\n\n", $response, 2);
		else:
			$headers = $response;
			$content = '';
		endif;

		// turn the headers into an array like $http_response_header
		$headers = explode ("\n", $headers);
		foreach ($headers as $key => $header):
			if ($key):
				if ($header[0] === "\x20" or $header[0] === "\t"):
					$headers[$prev] .= " " . trim ($header);
					unset ($headers[$key]);
					continue;
				endif;
			endif;
			$prev = $key;
		endforeach;

	endif;

	return [trim ($content), $headers];
}


/* Checks the robots.txt file for a given URL and returns an array of disallowed
 * base URLs. Results are cached for the execution of this script. It matches
 * http://robotstxt.org/ roughly, not the more sophisticated 2019 standard draft
 * https://tools.ietf.org/html/draft-koster-rep-00. In particular, this method
 * ignores ALLOW fields and $ and * are not treated specially in URLs. # is
 * recognized as a comment though and redirects are followed 5 times.
 */
public static function fetch_robots_txt (
	$for_url
):	array
{
	static $robotstxt = [];
	$base = word::normalize_url ('/', $for_url);

	// return cached results
	if (isset ($robotstxt[$base])):
		return $robotstxt[$base];
	endif;

	$handle = curl_init ($base . 'robots.txt');
	curl_setopt ($handle, CURLOPT_FAILONERROR, false);
	curl_setopt ($handle, CURLOPT_FOLLOWLOCATION, true);
	curl_setopt ($handle, CURLOPT_HEADER, false);
	curl_setopt ($handle, CURLOPT_MAXREDIRS, 5);
	curl_setopt ($handle, CURLOPT_NOBODY, false);
	curl_setopt ($handle, CURLOPT_RETURNTRANSFER, true);
	curl_setopt ($handle, CURLOPT_TIMEOUT, self::timeout);
	curl_setopt ($handle, CURLOPT_USERAGENT, self::user_agent);
	$body = curl_exec ($handle);
	$errno = curl_errno ($handle);
	$status = (int) curl_getinfo ($handle, CURLINFO_HTTP_CODE);
	$type = (string) curl_getinfo ($handle, CURLINFO_CONTENT_TYPE);
	curl_close ($handle);

	// no success in retrieving the document is interpreted as no restriction
	$type = strtolower (trim (substr ($type, 0, strcspn ($type, ';'))));
	if ($errno or $status !== 200 or ($type !== '' && $type !== 'text/plain')):
		return $robotstxt[$base] = [];
	endif;
	
	// truncate the body
	// Note: consider terminating the transfer after this many bytes instead
	$body = substr ($body, 0, self::maxsize_robots);

	// $useragent will be set to 0 when in a group with disallow rules that
	// apply for all bots and to 1 specifically in group for this mouse bot
	$useragent = NULL;
	$disallowed = [0 => [], 1 => NULL];

	// analyze the robots.txt line by line
	foreach (explode ("\n", str_replace (array ("\r\n", "\r"), "\n", $body)) as $line):

		// an empty line indicates the end of a group for a user agent
		$line = trim ($line);
		if ($line === ''):
			$useragent = NULL;
			continue;
		endif;

		// skip full-line comments and remove in-line comments
		if ($line[0] === '#'):
			continue;
		elseif ($hash = strpos ($line, '#')):
			$line = substr ($line, 0, $hash);
		endif;			

		// separate field and value; ignore lines which don't have those
		if ($colon = strpos ($line, ':')):
			list ($field, $value) = explode (':', $line, 2);
			$field = strtolower ($field);
			$value = ltrim ($value);
		else:
			continue;
		endif;

		// recognize the fields user-agent and disallow; disallow rules are only
		// remembered when user-agent identifies this bot specifically or is *
		// for all bots
		if ($field === 'user-agent'):
			if ($value === '*'):
				$useragent = 0;
			elseif (substr (strtolower ($value), 0, 5) === 'mouse'):
				$useragent = 1;
				$disallowed[1] ??= [];
			endif;
		elseif ($field === 'disallow' and isset ($useragent) and $value !== ''):
			$disallowed[$useragent][] = $value;
		endif;
			
	endforeach;

	// return normalized forbidden URLs for MOUSE or, if not available, for all
	// user agents
	$urls = [];
	foreach ($disallowed[1] ?? $disallowed[0] as $url):
		$urls[] = word::normalize_url ($url, $base);
	endforeach;
	return $robotstxt[$base] = array_values (array_unique ($urls));
}


/* Normalizes $urls and returns an array of those among them which are in scope
 * of this search engine, i.e. those who start the right way (e.g. scheme,
 * domain) and whose paths do not have file extensions which are known not to be
 * indexable.
 */
private function filter_urls (
	array $urls
):	array
{
	foreach ($urls as $key => $url):
		$urls[$key] = word::normalize_url ($url, $this->base);
	endforeach;

	$urls = array_unique ($urls);

	foreach ($urls as $key => $url):
		if (!$this->is_in_scope ($url) or strlen ($url) > self::maxsize_url):
			unset ($urls[$key]);
		endif;
	endforeach;

	return array_values ($urls);
}


/* Returns whether an $url is in scope of the application. The scope is set when
 * initializing this document in the form of strings that the $url may start
 * with, usually scheme + host, and further restricted by file extensions that
 * are known not to be used with indexable documents.
 */
private function is_in_scope (
	string $url
):	bool
{
	// return false if a file extensions hints at an unacceptable mime type
	if (
		is_string ($path = parse_url ($url, PHP_URL_PATH))
		and $pos = strrpos ($path, '.')
		and in_array (strtolower (substr ($path, $pos + 1)), self::avoid_extensions, true)
	):
		return false;
	endif;

	// return true iff the URL matches an acceptable prefix
	foreach ($this->prefixes as $prefix):
		if (substr ($url, 0, $prefix[1]) === $prefix[0]):
			return true;
		endif;
	endforeach;

	return false;
}


/* Parses a HTML or XHTML file to extract links, title, description, keywords,
 * the text content and looks for signs that the document shall not be indexed.
 * The UTF-8 conversion is implicitely done by PHP's DomDocument for (X)HTML
 * documents.
 */
private function parse_html (
	string $content
):	void
{
	// load DOM tree for HTML files; the @ prevents PHP warnings in case of HMTL
	// markup errors in the document
	$dom = new \DOMDocument ();
	@$dom->loadHTML ($content);

	// reset base url, if <base href='...'> includes an absolute URL
	foreach ($dom->getElementsByTagName ('base') as $node):
		if ($href = $node->getAttribute ('href')):
			$href = trim ($href);
			if (parse_url ($href, PHP_URL_SCHEME) and parse_url ($href, PHP_URL_HOST)):
				$this->base = word::strip_fragment ($href);
			endif;
			break;
		endif;
	endforeach;

	// get author, description, keywords and some more from meta data
	$authors = $keywords = [];
	foreach ($dom->getElementsByTagName ('meta') as $node):
		$name = strtolower (trim ($node->getAttribute ('name')));
		if ($name === 'author'):
			$authors[] = trim ($node->getAttribute ('content'));
		elseif ($name === 'description'):
			$this->description = mb_substr (word::collapse_whitespace ($node->getAttribute ('content')), 0, self::maxsize_snippet, 'UTF-8');
		elseif ($name === 'generator'):
			$generator = trim ($node->getAttribute ('content'));
		elseif ($name === 'keywords'):
			$values = array_map ('trim', explode (',', $node->getAttribute ('content')));
			array_push ($keywords, ...$values);
		elseif ($name === 'robots'):
			foreach (explode (',', strtolower ($node->getAttribute ('content'))) as $value):
				$v = explode (':', $value, 2);
				$this->robots[trim ($v[0])] = trim ($v[1] ?? $v[0]);
			endforeach;
		endif;
	endforeach;
	$this->author = mb_substr (implode (', ', $authors), 0, self::maxsize_author, 'UTF-8');
	$this->keywords = mb_substr (implode (', ', array_diff (array_unique ($keywords), [''])), 0, self::maxsize_keywords, 'UTF-8');

	// Collect different types of links from the Location header, usually
	// available for 3xx and 201 status codes, and from the HTML unless told not
	// to by a robots-meta-tag with nofollow attribute.
	$links = [];
	if (!isset ($this->robots['nofollow']) and !isset ($this->robots['none'])):
		if (isset ($this->headers['location'])):
			$links[] = $this->headers['location'];
		endif;
		foreach ($dom->getElementsByTagName ('a') as $node):
			if (!array_intersect (word::tokens ($node->getAttribute ('rel')), self::avoid_link_relations)):
				$href = $node->getAttribute ('href') and $links[] = $href;
			endif;
		endforeach;
		foreach ($dom->getElementsByTagName ('link') as $node):
			if ($node->hasAttribute ('type')):
				if (!in_array (strtolower ($node->getAttribute ('type')), self::accept_types, true)):
					continue;
				endif;
			endif;
			if (!array_intersect (word::tokens ($node->getAttribute ('rel')), self::avoid_link_relations)):
				$href = $node->getAttribute ('href') and $links[] = $href;
			endif;
		endforeach;
		foreach ($dom->getElementsByTagName ('frame') as $node):
			$href = $node->getAttribute ('src') and $links[] = $href;
		endforeach;
		foreach ($dom->getElementsByTagName ('iframe') as $node):
			$href = $node->getAttribute ('src') and $links[] = $href;
		endforeach;
	endif;
	$this->links = $this->filter_urls ($links);

	// A HTML document with a <meta name='robots'> tag whose content attribute
	// contains the string 'noindex' is not indexable. We can stop the analysis.
	if (isset ($this->robots['noindex']) or isset ($this->robots['none'])):
		$this->indexable = false;
		return;
	endif;

	// A HTML document that defines a canonical URL that is different from this
	// document's URL but is also in scope of the application is not indexable.
	// The canonical URL shall be indexed instead of this URL. Stop analysis.
	foreach ($dom->getElementsByTagName ('link') as $node):
		if ($rel = $node->getAttribute ('rel')):
			$url = word::normalize_url ($node->getAttribute ('href'), $this->base);
			if (stripos ($rel, 'canonical') !== false and $url !== $this->url and $this->is_in_scope ($url)):
				$this->indexable = false;
				return;
			endif;
		endif;
	endforeach;

	// We are going to manipulate the DOM now to prepare for text extraction.

	// remove irrelevant, not longer accurate or uncharacteristic elements
	foreach (self::elements_to_remove as $name):
		foreach (iterator_to_array ($dom->getElementsByTagName ($name)) as $node):
			$node->parentNode->removeChild ($node);
		endforeach;
	endforeach;

	// remove further elements known to be uncharacteristic for an Aneamal page
	if (isset ($generator) and stripos ($generator, 'Aneamal') !== false):
		foreach (self::elements_to_remove_from_nml as $name):
			foreach (iterator_to_array ($dom->getElementsByTagName ($name)) as $node):
				$node->parentNode->removeChild ($node);
			endforeach;
		endforeach;
	endif;

	// remove hidden elements and those whose roles like role=navigation signal
	// that the element is usually not relevant to this document but OTHER
	// documents or is just repetitive
	foreach (iterator_to_array ($dom->getElementsByTagName ('*')) as $node):
		if ($node->hasAttribute ('hidden')):
			$node->parentNode->removeChild ($node);
		elseif ($node->hasAttribute ('role')):
			if (array_intersect (word::tokens ($node->getAttribute ('role')), self::elements_with_roles_to_remove)):
				$node->parentNode->removeChild ($node);
			endif;
		endif;
	endforeach;

	// Save the first <title> element, which contains the document title, as
	// $this->title and strip it from the DOM. We do not remove other title
	// elements, as they are useful elements in embedded SVG code for example.
	foreach (iterator_to_array ($dom->getElementsByTagName ('title')) as $node):
		isset ($node->textContent) and $this->title = mb_substr (word::collapse_whitespace ($node->textContent), 0, self::maxsize_title, 'UTF-8');
		$node->parentNode->removeChild ($node);
		break;
	endforeach;

	// replace elements that separate content with space to prevent words being
	// merged when we extract textContent later
	foreach (self::elements_to_replace_with_space as $name):
		foreach (iterator_to_array ($dom->getElementsByTagName ($name)) as $node):
			$node->parentNode->replaceChild ($dom->createTextNode ("\x20"), $node);
		endforeach;
	endforeach;

	// <img> elements are replaced by their alt attribute, which holds a textual
	// alternative of the image, padded with space.
	foreach (iterator_to_array ($dom->getElementsByTagName ('img')) as $node):
		$node->parentNode->replaceChild ($dom->createTextNode ("\x20" . $node->getAttribute ('alt') . "\x20"), $node);
	endforeach;

	// pad block elements that can contain text with white space, so that words
	// do not get merged when we extract textContent later
	foreach (self::elements_to_pad as $name):
		foreach (iterator_to_array ($dom->getElementsByTagName ($name)) as $node):
			$node->appendChild ($dom->createTextNode ("\x20"));
			$node->insertBefore ($dom->createTextNode ("\x20"), $node->firstChild);
		endforeach;
	endforeach;

	// extract the remaining textual content
	$this->text = is_object ($dom->documentElement)? trim ($dom->documentElement->textContent): '';

	// title or textual content must be present for the document to be indexed
	if ($this->text === '' and $this->title === ''):
		$this->indexable = false;
	endif;

	// End here, if snippet use has been forbidden for robots. Otherwise use an
	// existing description as snippet. If none exists, generate a snippet
	// from paragraphs joined with " … ".
	if (isset ($this->robots['nosnippet'])):
		return;
	elseif ($this->description !== ''):
		$this->snippet = $this->description;
	else:
		foreach (iterator_to_array ($dom->getElementsByTagName ('*')) as $node):
			if ($node->hasAttribute ('data-nosnippet')):
				$node->parentNode->removeChild ($node);
			endif;
		endforeach;
		$desc = '';
		foreach (iterator_to_array ($dom->getElementsByTagName ('p')) as $node):
			$p = word::collapse_whitespace ($node->textContent);
			if ($desc === ''):
				$desc = $p;
			elseif (substr ($desc, -3) === "\xE2\x80\xA6" or substr ($p, 0, 3) === "\xE2\x80\xA6"):
				$desc .= " $p";
			else:
				$desc .= " \xE2\x80\xA6 $p";
			endif;
			if (mb_strlen ($desc, 'UTF-8') >= self::maxsize_snippet):
				break;
			endif;
		endforeach;
		if (mb_strlen ($desc, 'UTF-8') > self::maxsize_snippet):
			$this->snippet = mb_substr ($desc, 0, self::maxsize_snippet - 1, 'UTF-8') . "\xE2\x80\xA6";
		else:
			$this->snippet = $desc;
		endif;
	endif;
}


/* There is not much to parse about plain text, so we just report the link from
 * the location header in the unlikely case that there is one, convert the
 * content to UTF-8 and save a snippet.
 */
private function parse_plain_text (
	string $content
):	void
{
	if (isset ($this->headers['location']) and !isset ($this->robots['nofollow']) and !isset ($this->robots['none'])):
		$this->links = $this->filter_urls ([$this->headers['location']]);
	endif;

	// stop the analysis of texts that are not indexable
	if ($this->indexable === false):
		return;
	endif;

	$charset = $this->get_charset_from_header ();
	$this->text = word::utf8 ($content, $charset? $charset: 'UTF-8');

	if (!isset ($this->robots['nosnippet'])):
		$this->snippet = mb_substr (word::collapse_whitespace ($this->text), 0, self::maxsize_snippet, 'UTF-8');
	endif;
}

} // class retriever

// The End