<?php // Please do not edit this file. If you do, save as UTF-8 without BOM!

/* Copyright 2019-2022 Martin Janecke <mouse@prlbr.de>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* This module integrates a search form into an Aneamal page. It comes with and
 * triggers the crawler which builds the SQLite database to search in. The
 * database and further files are saved in the /aneamal/cache. Example use:
 *
 * [x-mouse:Find it!]->https://example.org/
 */

// declare (strict_types = 1); // only during development

namespace prlbr\mouse;

// error_reporting (E_ALL); // only during development

require_once __DIR__ . '/assets/finder.php';
require_once __DIR__ . '/assets/functions.php';
require_once __DIR__ . '/assets/i18n.php';


/* Return a closure which will be called for every instance of the search
 * module. $_ will be filled with data passed from the Aneamal core to this
 * module. $max on the other hand will not receive any value; on the contrary,
 * the declaration $max = 99 tells the Aneamal core that this module accepts
 * up to ninety-nine links to sites that shall be indexed.
 */
return function (
	array $_,
	int $max = 99
):	string
{
	// identificator for the current instance of the search module on a page
	// so that the correct instance responds to an HTTP post query
	static $instance = 0;
	$name = 'mouse-' . strval (++$instance);

	// interpret @x-mouse:snippet=... parameters such as the number of snippet
	// characters to be displayed and results per page in case of pagination
	$parameters = parameters ($_['meta']);
	$snippetlen = intval ($parameters['snippet'] ?? 0);
	$hitsperpage = intval ($parameters['page'] ?? PHP_INT_MAX);

	// whether the crawler shall be triggered; set to true later
	$trigger_crawler = false;

	// check all links whether they are absolute URLs; suggest an absoulte URL
	// to the homepage in case they are not
	foreach ($_['links'] as $i => $link):
		if (
			$scheme = parse_url ($link, PHP_URL_SCHEME)
			and in_array (strtolower ($scheme), ['http', 'https'], true)
			and parse_url ($link, PHP_URL_HOST)
		):
			continue;
		elseif (
			isset ($_SERVER['REQUEST_SCHEME'])
			and in_array (strtolower ($_SERVER['REQUEST_SCHEME']), ['http', 'https'], true)
			and isset ($_SERVER['HTTP_HOST'])
		):
			$home = $_SERVER['REQUEST_SCHEME'] . '://' . $_SERVER['HTTP_HOST'] . $_['home'] . '/';
			return message ('Search engine links must be absolute URLs like "' . $home . '", but "' . $link . '" is not.');
		else:
			return message ('Search engine links must be absolute URLs, but "' . $link . '" is not.');
		endif;
	endforeach;

	// start preparing the output
	$out = ["<div class='mouse'>"];

	// create a short hash to identify the realm of the search module that is
	// defined by the links
	$hash = realm ($_['links']);

	$cache = "$_[root]/aneamal/private/mouse/$hash";
	$taskfile = "$cache/task.txt";
	$donefile = "$cache/done.txt";
	$dbpattern = "$cache/*.db";

	// create a task file for the current search realm, if it does not exist yet
	if (!file_exists ($taskfile)):
		if (!is_dir ($cache) and !@mkdir ($cache, 0777, true)):
			return message ('Make "/aneamal/private/mouse" writable for the search engine to work.');
		elseif (!is_writable ($cache)):
			return message ('Make "/aneamal/private/mouse/' . $hash .  '" writable for the search engine to work.');
		elseif (!@file_put_contents ($taskfile, implode ("\t", $_['links']) . "\n" . implode ("\n", $_['links']))):
			return message ('Setting up the search engine failed for an unknown reason.');
		endif;
		// create the donefile and date it back, so that the crawler will not
		// wait before analyzing the first page
		@touch ($donefile, time () - 30);

		$trigger_crawler = true;
	endif;

	// ARIA label for the search button
	$button = esc ($_['clue'] ?? i18n ($_['lang'], 'Search'));
	$aria = $button === ''? '': " aria-label='$button'";

	// the search term submitted by a user
	$term = isset ($_POST[$name])? trim ($_POST[$name]): '';
	$value = esc ($term);

	// Generate the HTML code for the search form. The default button text is
	// U+1F50D LEFT-POINTING MAGNIFYING GLASS plus U+FE0E VARIATION SELECTOR-15.
	// Apparently some browsers/operating systems ignore the text variation
	// selector though and display the magnifying glass as emoji anyway.
	$out[] = "<form id='$name-form' action='#mouse' method='post' role='search' class='mouse-form'>";
	$out[] = "<input name='$name' type='search' maxlength='2048' value='$value' required><button type='submit'$aria><span>\xF0\x9F\x94\x8D\xEF\xB8\x8E</span></button>";
	$out[] = "</form>";

	// show a help page, if only a question mark is entered as search term ...
	if ($term === '?'):

		// prepare a message with the URL of the crawler
		$out[] = help ($_['here'], $_['home'], $hash);
		$trigger_crawler = true;

	// ... or do an actual search, if something else is entered
	elseif ($term !== ''):

		// The directory needs to be writable for SQLite to work in WAL mode.
		if (!is_writable ($cache)):
			return message ('Make "/aneamal/private/mouse/' . $hash .  '" writable for the search engine to work.');
		endif;

		// Choose the second database if two exist, the only one if one exists,
		// and try to use a legacy database if no current one exists.
		if ($dbs = glob ($dbpattern)):
			$search = new finder ($dbs[max (0, count ($dbs) - 2)]);
			$hits = $search->find ($term);
		elseif ($dbs = glob ($_['root'] . '/aneamal/cache/mouse-' . $hash . '-*.db')):
			$search = new finder ($dbs[max (0, count ($dbs) - 2)]);
			$hits = $search->find ($term);
		else:
			$hits = [];
		endif;
		$n = count ($hits);

		// BEGIN pagination
		if ($n > $hitsperpage and $hitsperpage > 0):
			// calculate the number of pages with search results
			$pages = intdiv ($n, $hitsperpage) + ($n % $hitsperpage? 1: 0);
			// determine the current page number, starting at 1
			$page = min (max (intval ($_POST['page'] ?? 1), 1), $pages);
			// calculate the first search result on this page, all starting at 1
			$start = $hitsperpage * ($page - 1) + 1;
			// reduce the search results to those on the current page
			$hits = array_slice ($hits, $start - 1, $hitsperpage);
			// generate HTML buttons to navigate to each search result page
			$pagination[] = "<nav>";
			for ($p = 1; $p <= $pages; ++$p):
				// buttons are labeled with the search result range on the page
				if ($p === $pages && $n % $hitsperpage === 1):
					$label = "$n";
				else:
					$label = strval ($s = $hitsperpage * ($p - 1) + 1) . "\xE2\x80\x93" . strval (min ($s + $hitsperpage - 1, $n));
				endif;
				$current = $p === $page? " aria-current='true'": '';
				$pagination[] = "<button form='$name-form' type='submit' name='page' value='$p'$current><span>$label</span></button>";
			endfor;
			$pagination = implode ("\n", $pagination) . "\n</nav>\n";
		else:
			$start = 1;
			$pagination = '';
		endif;
		// END pagination

		// tell how many search results there are before listing the actual
		// results so that especially blind users know early what to expect
		$out[] = "<div id='mouse' class='mouse-results' role='region' aria-labelled-by='mouse-count'>";
		$out[] = "<p id='mouse-count'>" . sprintf (i18n ($_['lang'], '%1$d search results for <i>%2$s</i>', $n), $n, esc ($term)) . '</p>';

		// turn the results into HTML for output
		if (!empty ($hits)):
			$out[] = "<ol class='mouse-list' start='$start'>";
			foreach ($hits as $result):
				if (isset ($result['score'])):
					$attribute = "data-score='" . esc ((string) $result['score']) . "'";
				else:
					$attribute = '';
				endif;
				$out[] = item ($result, $snippetlen, $attribute);
			endforeach;
			
			$out[] = '</ol>';
		endif;
		$out[] = "$pagination</div>";
		$trigger_crawler = true;

	endif;

	if ($trigger_crawler):
		$out[] = trigger_crawler ($_['here'], $hash);
	endif;
	$out[] = '</div>';

	return implode ("\n", $out);
};

// The End